/*
  specific_S12232ZA.cpp
  
  This file is part of Monochrome Graphic LCD library (MGLCD) for Arduino
  Uno/Mega 2560.
  
  MGLCD was developed by Synapse(Hiroshi Tanigawa) from 2012 to 2013.  This Library is
  originally distributed at Team Schaft's Homepage. 
  <http://www3.big.or.jp/~schaft/hardware>

  MGLCD is now under beta testing, so specification may be changed 
  in the future.

  MGLCD is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 2.1 of the License, or
  (at your option) any later version.

  MGLCD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with MGLCD.  If not, see <http://www.gnu.org/licenses/>.
  
*/

#include <avr/pgmspace.h>
#include "mglcd.h"

#if MGLCD_VERSION != 32
#error versions of this file and header file are different.
#endif

// Commands
#define SG12232_STATIC_DRIVE_ON_COMMAND   0xa5
#define SG12232_STATIC_DRIVE_OFF_COMMAND  0xa4
#define SG12232_RESET_COMMAND             0xe2

// Set one bit of I/O port. Operands must be constants.
static void inline sbi(volatile uint8_t *port, int bit)
{
  __asm__ volatile("sbi %0,%1 \n\t" :: "M"(port-0x20) ,"M"(bit));
} // sbi

// Clear one bit of I/O port. Operands must be constants.
static void inline cbi(volatile uint8_t *port, int bit)
{
  __asm__ volatile("cbi %0,%1\n\t" :: "M"(port-0x20), "M"(bit));
} // cbi
static inline void PortOut(const TPortBit &pb, unsigned char val)
{
    uint8_t oldSREG;
    
    oldSREG=SREG;
    cli();
    
    if(val) {
      *pb.OutReg|= pb.mask;
    } else {
      *pb.OutReg&=~pb.mask;
    } // if
    
    SREG=oldSREG;      
} // PortOut

static unsigned char GeneralReadData(unsigned char ctrl, const TPinSet &PinSet)
{
  unsigned char result;
  
  // Control A0 pin 
  PortOut(PinSet.A0_DIPin,ctrl&MGLCD_DATA);
  
  // Control R/W pin
  PortOut(PinSet.RWPin,1);
  
  // Control E1/E2 pin
  if(ctrl&MGLCD_CS1) {
    PortOut(PinSet.CS1_E1Pin,1);
  } else {
    PortOut(PinSet.CS2_E2Pin,1);
  }
  
  // read data
  result= (*PinSet.LcdBus.InReg[0]&PinSet.LcdBus.mask[0] ? 0x01 : 0)
         +(*PinSet.LcdBus.InReg[1]&PinSet.LcdBus.mask[1] ? 0x02 : 0)
         +(*PinSet.LcdBus.InReg[2]&PinSet.LcdBus.mask[2] ? 0x04 : 0)
         +(*PinSet.LcdBus.InReg[3]&PinSet.LcdBus.mask[3] ? 0x08 : 0)
         +(*PinSet.LcdBus.InReg[4]&PinSet.LcdBus.mask[4] ? 0x10 : 0)
         +(*PinSet.LcdBus.InReg[5]&PinSet.LcdBus.mask[5] ? 0x20 : 0)
         +(*PinSet.LcdBus.InReg[6]&PinSet.LcdBus.mask[6] ? 0x40 : 0)
         +(*PinSet.LcdBus.InReg[7]&PinSet.LcdBus.mask[7] ? 0x80 : 0);
         
  // Control E1/E2 pin
  if(ctrl&MGLCD_CS1) {
    PortOut(PinSet.CS1_E1Pin,0);
  } else {
    PortOut(PinSet.CS2_E2Pin,0);
  }

  return result;  
} // GeneralReadData

static unsigned char StandardReadData(unsigned char ctrl, const TPinSet &PinSet)
{
  unsigned char result;

  // Control A0 pin
  if(ctrl&MGLCD_DATA) {
    sbi(&STANDARD_A0_DI_PIN_OUT_REG,STANDARD_A0_DI_PIN_BIT);	
  } else {
    cbi(&STANDARD_A0_DI_PIN_OUT_REG,STANDARD_A0_DI_PIN_BIT);	
  } // if
  
  // Control R/W pin
  sbi(&STANDARD_RW_PIN_OUT_REG,STANDARD_RW_PIN_BIT);
  
  // control E1/E2 pin
  if(ctrl&MGLCD_CS1) {
    sbi(&STANDARD_CS1_E1_PIN_OUT_REG,STANDARD_CS1_E1_PIN_BIT);
  } else {
    sbi(&STANDARD_CS2_E2_PIN_OUT_REG,STANDARD_CS2_E2_PIN_BIT);
  } // if
  
  __asm__ volatile("nop\n\t"); // to meet access time
  __asm__ volatile("nop\n\t"); // to meet access time
    
    // read data
  result=(PINB & 0x0f)|(PIND & 0xf0);
    
  // control E1/E2 pin
  cbi(&STANDARD_CS1_E1_PIN_OUT_REG,STANDARD_CS1_E1_PIN_BIT);
  cbi(&STANDARD_CS2_E2_PIN_OUT_REG,STANDARD_CS2_E2_PIN_BIT);
  
  return result;
} // StandardReadData

static void GeneralWriteData(unsigned char ctrl, unsigned char data, const TPinSet &PinSet)
{
  uint8_t oldSREG;
    
  // Control A0 pin 
  PortOut(PinSet.A0_DIPin,ctrl&MGLCD_DATA); 

  // Control R/W pin
  PortOut(PinSet.RWPin,0);
  
  // save SREG and disable interrupt    
  oldSREG=SREG;
  cli();
    
  // set bus output mode
#if PORT_NUM==3
  switch(PinSet.LcdBus.PortNum) {
  case 3:
    *PinSet.LcdBus.NModeReg[2]|=PinSet.LcdBus.RegMask[2];
  case 2:
    *PinSet.LcdBus.NModeReg[1]|=PinSet.LcdBus.RegMask[1];
  case 1:
    *PinSet.LcdBus.NModeReg[0]|=PinSet.LcdBus.RegMask[0];
  } // switch
#else
  for(int i=0; i<PinSet.LcdBus.PortNum; i++) {
    *PinSet.LcdBus.NModeReg[i]|=PinSet.LcdBus.RegMask[i];
  } // for i
#endif

  // send data on bus
  if(data&0x01) {
      *PinSet.LcdBus.OutReg[0]|= PinSet.LcdBus.mask[0];
  } else {
      *PinSet.LcdBus.OutReg[0]&=~PinSet.LcdBus.mask[0];
  } // if
  if(data&0x02) {
      *PinSet.LcdBus.OutReg[1]|= PinSet.LcdBus.mask[1];
  } else {
      *PinSet.LcdBus.OutReg[1]&=~PinSet.LcdBus.mask[1];
  } // if
  if(data&0x04) {
      *PinSet.LcdBus.OutReg[2]|= PinSet.LcdBus.mask[2];
  } else {
      *PinSet.LcdBus.OutReg[2]&=~PinSet.LcdBus.mask[2];
  } // if
  if(data&0x08) {
      *PinSet.LcdBus.OutReg[3]|= PinSet.LcdBus.mask[3];
  } else {
      *PinSet.LcdBus.OutReg[3]&=~PinSet.LcdBus.mask[3];
  } // if
  if(data&0x10) {
      *PinSet.LcdBus.OutReg[4]|= PinSet.LcdBus.mask[4];
  } else {
      *PinSet.LcdBus.OutReg[4]&=~PinSet.LcdBus.mask[4];
  } // if
  if(data&0x20) {
      *PinSet.LcdBus.OutReg[5]|= PinSet.LcdBus.mask[5];
  } else {
      *PinSet.LcdBus.OutReg[5]&=~PinSet.LcdBus.mask[5];
  } // if
  if(data&0x40) {
      *PinSet.LcdBus.OutReg[6]|= PinSet.LcdBus.mask[6];
  } else {
      *PinSet.LcdBus.OutReg[6]&=~PinSet.LcdBus.mask[6];
  } // if
  if(data&0x80) {
      *PinSet.LcdBus.OutReg[7]|= PinSet.LcdBus.mask[7];
  } else {
      *PinSet.LcdBus.OutReg[7]&=~PinSet.LcdBus.mask[7];
  } // if
    
  // control E1 pin
  if(ctrl&MGLCD_CS1) {
    *PinSet.CS1_E1Pin.OutReg|= PinSet.CS1_E1Pin.mask;
    *PinSet.CS1_E1Pin.OutReg&=~PinSet.CS1_E1Pin.mask;
  } // if

  // control E2 pin
  if(ctrl&MGLCD_CS2) {
    *PinSet.CS2_E2Pin.OutReg|= PinSet.CS2_E2Pin.mask;
    *PinSet.CS2_E2Pin.OutReg&=~PinSet.CS2_E2Pin.mask;
  } // if

    // set bus pullup mode
#if PORT_NUM==3
  switch(PinSet.LcdBus.PortNum) {
  case 3:
    *PinSet.LcdBus.NModeReg[2]&=~PinSet.LcdBus.RegMask[2];
    *PinSet.LcdBus.NOutReg [2]|= PinSet.LcdBus.RegMask[2];
  case 2:
    *PinSet.LcdBus.NModeReg[1]&=~PinSet.LcdBus.RegMask[1];
    *PinSet.LcdBus.NOutReg [1]|= PinSet.LcdBus.RegMask[1];
  case 1:
    *PinSet.LcdBus.NModeReg[0]&=~PinSet.LcdBus.RegMask[0];
    *PinSet.LcdBus.NOutReg [0]|= PinSet.LcdBus.RegMask[0];
  } // switch
#else
  for(int i=0; i<PinSet.LcdBus.PortNum; i++) {
    *PinSet.LcdBus.NModeReg[i]&=~PinSet.LcdBus.RegMask[i];
    *PinSet.LcdBus.NOutReg [i]|= PinSet.LcdBus.RegMask[i];
  } // for i
#endif 

  // restore SREG and enable interrupt
  SREG=oldSREG;   
} // GeneralWriteData

static void StandardWriteData(unsigned char ctrl, unsigned char data, const TPinSet &PinSet)
{
  uint8_t oldSREG;
  
  // Control A0 pin
  if(ctrl&MGLCD_DATA) {
    sbi(&STANDARD_A0_DI_PIN_OUT_REG,STANDARD_A0_DI_PIN_BIT);
  } else {
    cbi(&STANDARD_A0_DI_PIN_OUT_REG,STANDARD_A0_DI_PIN_BIT);
  } // if
  
  // Control R/W pin
  cbi(&STANDARD_RW_PIN_OUT_REG,STANDARD_RW_PIN_BIT);
    
  // save SREG and disable interrupt
  oldSREG=SREG;
  cli();
    
  // set bus output mode
  DDRB|=0x0f;
  DDRD|=0xf0;

  // send data on bus
  PORTB=(PORTB & 0xf0)|(data & 0x0f);
  PORTD=(PORTD & 0x0f)|(data & 0xf0);

  // Control E1/E2 pin 
  if(ctrl&MGLCD_CS1) sbi(&STANDARD_CS1_E1_PIN_OUT_REG,STANDARD_CS1_E1_PIN_BIT);
  if(ctrl&MGLCD_CS2) sbi(&STANDARD_CS2_E2_PIN_OUT_REG,STANDARD_CS2_E2_PIN_BIT);
  __asm__ volatile("nop\n\t"); // to meet access time
  __asm__ volatile("nop\n\t"); // to meet access time
  __asm__ volatile("nop\n\t"); // to meet access time
  __asm__ volatile("nop\n\t"); // to meet access time
  __asm__ volatile("nop\n\t"); // to meet access time
  cbi(&STANDARD_CS1_E1_PIN_OUT_REG,STANDARD_CS1_E1_PIN_BIT);
  cbi(&STANDARD_CS2_E2_PIN_OUT_REG,STANDARD_CS2_E2_PIN_BIT);

  // set bus pullup mode
  DDRB &=~0x0f;
  PORTB|= 0x0f;
  DDRD &=~0xf0;    
  PORTD|= 0xf0;

  // restore SREG and enable interrupt
  SREG=oldSREG; 
} // StandardWriteData

unsigned char (*const specific_S12232ZA::GeneralReadData)(unsigned char ctrl, const TPinSet &PinSet) = ::GeneralReadData;
unsigned char (*const specific_S12232ZA::StandardReadData)(unsigned char ctrl, const TPinSet &PinSet) = ::StandardReadData;
void (*const specific_S12232ZA::GeneralWriteData )(unsigned char ctrl, unsigned char data, const TPinSet &PinSet) = ::GeneralWriteData;
void (*const specific_S12232ZA::StandardWriteData)(unsigned char ctrl, unsigned char data, const TPinSet &PinSet) = ::StandardWriteData;

void specific_S12232ZA::Reset(void)
{
  PortOut(PinSet.A0_DIPin ,1);
  PortOut(PinSet.CS1_E1Pin,0);
  PortOut(PinSet.CS2_E2Pin,0);
  PortOut(PinSet.EPin     ,0);
  PortOut(PinSet.RWPin    ,0);
  WriteData(MGLCD_CS1_CS2_COMMAND, SG12232_RESET_COMMAND,PinSet);
  while(ReadData(MGLCD_CS1_COMMAND,PinSet)&0x90 || ReadData(MGLCD_CS2_COMMAND,PinSet)&0x90); // wait until LCD is ready
  WriteData(MGLCD_CS1_CS2_COMMAND, SG12232_STATIC_DRIVE_OFF_COMMAND,PinSet);
  WriteData(MGLCD_CS1_CS2_COMMAND, DisplayOnCommand                ,PinSet);
  WriteData(MGLCD_CS1_CS2_COMMAND, DisplayLineCommand              ,PinSet);
  WriteData(MGLCD_CS1_CS2_COMMAND, PageAddressCommand              ,PinSet);
  WriteData(MGLCD_CS1_CS2_COMMAND, ColumnAddressCommand            ,PinSet);
} // specific_S12232ZA::Reset
